==> klayout-0.15.tar.gz: klayout-0.15/src/tlVariant.h <==

/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#ifndef _HDR_tlVariant
#define _HDR_tlVariant

#include <string>
#include <vector>
#include <stdexcept>

#include "tlInternational.h"
#include "tlException.h"
#include "tlAssert.h"

namespace tl
{

/**
 *  @brief A base class whose derivatives encapsulate a user type
 */
class VariantUserBase
{
public:
  VariantUserBase () { }
  virtual ~VariantUserBase () { }

  virtual bool equal (const VariantUserBase *) const = 0;
  virtual bool less (const VariantUserBase *) const = 0;
  virtual VariantUserBase *clone () const = 0;
};

/**
 *  @brief A derived class encapsulating a certain user type
 */
template <class T>
class VariantUser
  : public VariantUserBase
{
public:
  VariantUser (const T &t) : mp_t (new T (t)) { }
  ~VariantUser () { delete mp_t; mp_t = 0; }

  //  TODO: not implemented to be more generic
  virtual bool equal (const VariantUserBase *) const { tl_assert (false); return false; }
  virtual bool less (const VariantUserBase *) const { tl_assert (false); return false; }

  virtual VariantUserBase *clone () const { return new VariantUser (*mp_t); }

  T &get () { return *mp_t; }
  const T &get () const { return *mp_t; }

private:
  T *mp_t;
};

/**
 *  @brief A derived class encapsulating a certain user type that has the ability to compare for equality
 */
template <class T>
class VariantUserWithEqual
  : public VariantUser<T>
{
public:
  VariantUserWithEqual (const T &t) : VariantUser<T> (t) { }

  virtual bool equal (const VariantUserBase *d) const 
  { 
    const VariantUserWithEqual<T> *dd = dynamic_cast <const VariantUserWithEqual<T> *> (d);
    return dd && this->get () == dd->get ();
  }
    
  virtual VariantUserBase *clone () const { return new VariantUserWithEqual (this->get ()); }
};

/**
 *  @brief A derived class encapsulating a certain user type that has the ability to compare for equality and has a "less" compare function
 */
template <class T>
class VariantUserWithEqualAndLess
  : public VariantUserWithEqual<T>
{
public:
  VariantUserWithEqualAndLess (const T &t) : VariantUserWithEqual<T> (t) { }

  virtual bool less (const VariantUserBase *d) const
  { 
    const VariantUserWithEqualAndLess<T> *dd = dynamic_cast <const VariantUserWithEqualAndLess<T> *> (d);
    return dd && this->get () < dd->get ();
  }

  virtual VariantUserBase *clone () const { return new VariantUserWithEqualAndLess (this->get ()); }
};


/**
 *   @brief A basic variant type
 *
 *   This variant is capable of storing long, double, std::string, void (nil) and lists
 *   of other variants.
 */
class Variant
{
public:
  enum type { t_nil, t_long, t_double, t_string, t_list, t_user };

  typedef std::vector<tl::Variant>::const_iterator iterator;

  /**
   *  @brief Tags to pass to the user object constructor
   *
   *  no_compare_tag: do not support comparison of user objects,
   *  has_equal_tag: user object supports operator==
   *  has_less_tag: user object supports operator== and operator<
   */
  struct no_compare_tag { };
  struct has_equal_tag { };
  struct has_less_tag { };

  /**
   *  @brief Initialize the Variant with "nil"
   */
  Variant ();

  /**
   *  @brief Initialize the Variant with "string"
   */
  Variant (const std::string &s);

  /**
   *  @brief Initialize the Variant with "string"
   */
  Variant (const char *s);

  /**
   *  @brief Initialize the Variant with "double"
   */
  Variant (double d);

  /**
   *  @brief Initialize the Variant with "long"
   */
  Variant (long l);

  /**
   *  @brief Initialize with a user type
   */
  template <class T>
  Variant (unsigned int utype, const T &t, no_compare_tag tag = no_compare_tag ())
    : m_type (type (t_user + utype)), m_string (0)
  {
    m_var.mp_user = new VariantUser<T> (t);
  }

  /**
   *  @brief Initialize with a user type
   */
  template <class T>
  Variant (unsigned int utype, const T &t, has_equal_tag)
    : m_type (type (t_user + utype)), m_string (0)
  {
    m_var.mp_user = new VariantUserWithEqual<T> (t);
  }

  /**
   *  @brief Initialize with a user type
   */
  template <class T>
  Variant (unsigned int utype, const T &t, has_less_tag)
    : m_type (type (t_user + utype)), m_string (0)
  {
    m_var.mp_user = new VariantUserWithEqualAndLess<T> (t);
  }

  /**
   *  @brief Initialize the Variant with a list
   */
  template <class Iter>
  Variant (Iter from, Iter to)
    : m_type (t_list), m_string (0)
  {
    m_var.m_list = new std::vector<tl::Variant> (from, to);
  }

  /**
   *  @brief Copy ctor
   */
  Variant (const Variant &v);

  /**
   *  @brief Destructor
   */
  ~Variant ();

  /**
   *  @brief Assignment
   */
  Variant &operator= (const Variant &v);

  /**
   *  @brief Assignment of a string
   */
  Variant &operator= (const char *v);

  /**
   *  @brief Assignment of a string
   */
  Variant &operator= (const std::string &v);

  /**
   *  @brief Assignment of a double
   */
  Variant &operator= (double d);

  /**
   *  @brief Assignment of a long
   */
  Variant &operator= (long l);

  /**  
   *  @brief Equality
   *
   *  For user types, this is not implemented yet.
   */
  bool operator== (const Variant &d) const;

  /**  
   *  @brief Inequality
   *
   *  For user types, this is not implemented yet.
   */
  bool operator!= (const Variant &d) const
  {
    return !operator== (d);
  }

  /**  
   *  @brief Comparison
   *
   *  For user types, this is not implemented yet.
   */
  bool operator< (const Variant &d) const;

  /**
   *  @brief Conversion to a string
   *
   *  This performs the conversion to a string as far as possible.
   *  No conversion is provided to user types currently.
   */
  const char *to_string () const;

  /**
   *  @brief Conversion to a long
   *
   *  This performs the conversion to a long as far as possible.
   *  No conversion is provided to user types currently.
   */
  long to_long () const;

  /**
   *  @brief Conversion to a double
   *
   *  This performs the conversion to a double value as far as possible.
   *  No conversion is provided to user types currently.
   */
  double to_double () const;

  /** 
   *  @brief Convert to the given user type
   */
  template <class T> 
  T &to_user ()
  {
    tl_assert (is_user ());
    VariantUser<T> *vu = dynamic_cast<VariantUser<T> *> (m_var.mp_user);
    tl_assert (vu != 0);
    return vu->get ();
  }

  /** 
   *  @brief Convert to the given user type (const version)
   */
  template <class T> 
  const T &to_user () const
  {
    tl_assert (is_user ());
    const VariantUser<T> *vu = dynamic_cast<const VariantUser<T> *> (m_var.mp_user);
    tl_assert (vu != 0);
    return vu->get ();
  }

  /**
   *  @brief Get the list, if it is one
   */
  const std::vector<tl::Variant> &get_list () const
  {
    tl_assert (m_type == t_list);
    return *m_var.m_list;
  }

  /**
   *  @brief Get the list iterators, if it is one
   */
  iterator begin () const
  {
    tl_assert (m_type == t_list);
    return m_var.m_list->begin ();
  }

  /**
   *  @brief Get the list iterators, if it is one
   */
  iterator end () const
  {
    tl_assert (m_type == t_list);
    return m_var.m_list->end ();
  }

  /**
   *  @brief Add a element to the list
   */
  void push (const tl::Variant &v)
  {
    tl_assert (m_type == t_list);
    m_var.m_list->push_back (v);
  }

  /**
   *  @brief Get the list, if it is one
   */
  std::vector<tl::Variant> &get_list ()
  {
    tl_assert (m_type == t_list);
    return *m_var.m_list;
  }

  /**
   *  @brief Test, if it is a double
   */
  bool is_double () const
  {
    return m_type == t_double;
  }

  /**
   *  @brief Test, if it is a long
   */
  bool is_long () const
  {
    return m_type == t_long;
  }

  /**
   *  @brief Test, if it is a string
   */
  bool is_string () const
  {
    return m_type == t_string;
  }

  /**
   *  @brief Test, if it is nil
   */
  bool is_nil () const
  {
    return m_type == t_nil;
  }

  /**
   *  @brief Test, if it is a list
   */
  bool is_list () const
  {
    return m_type == t_list;
  }

  /**
   *  @brief Test, if this is a user type
   */
  bool is_user () const
  {
    return m_type >= t_user;
  }

  /**
   *  @brief Get the user type if it is one
   */
  unsigned int user_type () const
  {
    tl_assert (is_user ());
    return (unsigned int) m_type - (unsigned int) t_user;
  }

  /**
   *  @brief A method to deliver an empty-list variant
   */
  static tl::Variant empty_list ();

  /**
   *  @brief Convert the Variant to a string that can be parsed with the Extractor
   *
   *  No conversion is provided for user types and "nil" currently.
   */
  std::string to_parsable_string () const;

private:
  type m_type;

  union {
    std::vector<tl::Variant> *m_list;
    double m_double;
    long m_long;
    VariantUserBase *mp_user;
  } m_var;

  //  this will hold the string if it is valid
  mutable char *m_string;

  void reset ();
};

/**
 *  @brief Special extractors for the variant
 */

class Extractor;
template <class T> void extractor_impl (tl::Extractor &ex, T &);
template <class T> bool test_extractor_impl (tl::Extractor &ex, T &);

template <> void extractor_impl (tl::Extractor &ex, tl::Variant &p);
template <> bool test_extractor_impl (tl::Extractor &ex, tl::Variant &p);

} // namespace tl

#endif

==> klayout-0.19.3.tar.gz: klayout-0.19.3/src/tlVariant.h <==

/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#ifndef _HDR_tlVariant
#define _HDR_tlVariant

#include <string>
#include <vector>
#include <stdexcept>

#include "tlInternational.h"
#include "tlException.h"
#include "tlAssert.h"

namespace tl
{

/**
 *  @brief A base class whose derivatives encapsulate a user type
 */
class VariantUserBase
{
public:
  VariantUserBase () { }
  virtual ~VariantUserBase () { }

  virtual bool equal (const VariantUserBase *) const = 0;
  virtual bool less (const VariantUserBase *) const = 0;
  virtual VariantUserBase *clone () const = 0;
};

/**
 *  @brief A derived class encapsulating a certain user type
 */
template <class T>
class VariantUser
  : public VariantUserBase
{
public:
  VariantUser (const T &t) : mp_t (new T (t)) { }
  ~VariantUser () { delete mp_t; mp_t = 0; }

  //  TODO: not implemented to be more generic
  virtual bool equal (const VariantUserBase *) const { tl_assert (false); return false; }
  virtual bool less (const VariantUserBase *) const { tl_assert (false); return false; }

  virtual VariantUserBase *clone () const { return new VariantUser (*mp_t); }

  T &get () { return *mp_t; }
  const T &get () const { return *mp_t; }

private:
  T *mp_t;
};

/**
 *  @brief A derived class encapsulating a certain user type that has the ability to compare for equality
 */
template <class T>
class VariantUserWithEqual
  : public VariantUser<T>
{
public:
  VariantUserWithEqual (const T &t) : VariantUser<T> (t) { }

  virtual bool equal (const VariantUserBase *d) const 
  { 
    const VariantUserWithEqual<T> *dd = dynamic_cast <const VariantUserWithEqual<T> *> (d);
    return dd && this->get () == dd->get ();
  }
    
  virtual VariantUserBase *clone () const { return new VariantUserWithEqual (this->get ()); }
};

/**
 *  @brief A derived class encapsulating a certain user type that has the ability to compare for equality and has a "less" compare function
 */
template <class T>
class VariantUserWithEqualAndLess
  : public VariantUserWithEqual<T>
{
public:
  VariantUserWithEqualAndLess (const T &t) : VariantUserWithEqual<T> (t) { }

  virtual bool less (const VariantUserBase *d) const
  { 
    const VariantUserWithEqualAndLess<T> *dd = dynamic_cast <const VariantUserWithEqualAndLess<T> *> (d);
    return dd && this->get () < dd->get ();
  }

  virtual VariantUserBase *clone () const { return new VariantUserWithEqualAndLess (this->get ()); }
};


/**
 *   @brief A basic variant type
 *
 *   This variant is capable of storing long, double, std::string, void (nil) and lists
 *   of other variants.
 */
class Variant
{
public:
  enum type { t_nil, t_long, t_id, t_double, t_string, t_list, t_user };

  typedef std::vector<tl::Variant>::const_iterator iterator;

  /**
   *  @brief Tags to pass to the user object constructor
   *
   *  no_compare_tag: do not support comparison of user objects,
   *  has_equal_tag: user object supports operator==
   *  has_less_tag: user object supports operator== and operator<
   */
  struct no_compare_tag { };
  struct has_equal_tag { };
  struct has_less_tag { };

  /**
   *  @brief Initialize the Variant with "nil"
   */
  Variant ();

  /**
   *  @brief Initialize the Variant with "string"
   */
  Variant (const std::string &s);

  /**
   *  @brief Initialize the Variant with "string"
   */
  Variant (const char *s);

  /**
   *  @brief Initialize the Variant with "double"
   */
  Variant (double d);

  /**
   *  @brief Initialize the Variant with "long"
   */
  Variant (long l);

  /**
   *  @brief Initialize the Variant with an "id" 
   *
   *  The "id" type is basically a long, but is supposed to be used as a representative for another value.
   *  One application for that type is a placeholder for an OASIS name until it is associated with a real value.
   */
  Variant (long l, bool /*dummy*/);

  /**
   *  @brief Initialize with a user type
   */
  template <class T>
  Variant (unsigned int utype, const T &t, no_compare_tag tag = no_compare_tag ())
    : m_type (type (t_user + utype)), m_string (0)
  {
    m_var.mp_user = new VariantUser<T> (t);
  }

  /**
   *  @brief Initialize with a user type
   */
  template <class T>
  Variant (unsigned int utype, const T &t, has_equal_tag)
    : m_type (type (t_user + utype)), m_string (0)
  {
    m_var.mp_user = new VariantUserWithEqual<T> (t);
  }

  /**
   *  @brief Initialize with a user type
   */
  template <class T>
  Variant (unsigned int utype, const T &t, has_less_tag)
    : m_type (type (t_user + utype)), m_string (0)
  {
    m_var.mp_user = new VariantUserWithEqualAndLess<T> (t);
  }

  /**
   *  @brief Initialize the Variant with a list
   */
  template <class Iter>
  Variant (Iter from, Iter to)
    : m_type (t_list), m_string (0)
  {
    m_var.m_list = new std::vector<tl::Variant> (from, to);
  }

  /**
   *  @brief Copy ctor
   */
  Variant (const Variant &v);

  /**
   *  @brief Destructor
   */
  ~Variant ();

  /**
   *  @brief Assignment
   */
  Variant &operator= (const Variant &v);

  /**
   *  @brief Assignment of a string
   */
  Variant &operator= (const char *v);

  /**
   *  @brief Assignment of a string
   */
  Variant &operator= (const std::string &v);

  /**
   *  @brief Assignment of a double
   */
  Variant &operator= (double d);

  /**
   *  @brief Assignment of a long
   */
  Variant &operator= (long l);

  /**  
   *  @brief Equality
   *
   *  For user types, this is not implemented yet.
   */
  bool operator== (const Variant &d) const;

  /**  
   *  @brief Inequality
   *
   *  For user types, this is not implemented yet.
   */
  bool operator!= (const Variant &d) const
  {
    return !operator== (d);
  }

  /**  
   *  @brief Comparison
   *
   *  For user types, this is not implemented yet.
   */
  bool operator< (const Variant &d) const;

  /**
   *  @brief Conversion to a string
   *
   *  This performs the conversion to a string as far as possible.
   *  No conversion is provided to user types currently.
   */
  const char *to_string () const;

  /**
   *  @brief Conversion to a long
   *
   *  This performs the conversion to a long as far as possible.
   *  No conversion is provided to user types currently.
   */
  long to_long () const;

  /**
   *  @brief Conversion to a double
   *
   *  This performs the conversion to a double value as far as possible.
   *  No conversion is provided to user types currently.
   */
  double to_double () const;

  /** 
   *  @brief Convert to the given user type
   */
  template <class T> 
  T &to_user ()
  {
    tl_assert (is_user ());
    VariantUser<T> *vu = dynamic_cast<VariantUser<T> *> (m_var.mp_user);
    tl_assert (vu != 0);
    return vu->get ();
  }

  /** 
   *  @brief Convert to the given user type (const version)
   */
  template <class T> 
  const T &to_user () const
  {
    tl_assert (is_user ());
    const VariantUser<T> *vu = dynamic_cast<const VariantUser<T> *> (m_var.mp_user);
    tl_assert (vu != 0);
    return vu->get ();
  }

  /**
   *  @brief Get the list, if it is one
   */
  const std::vector<tl::Variant> &get_list () const
  {
    tl_assert (m_type == t_list);
    return *m_var.m_list;
  }

  /**
   *  @brief Get the list iterators, if it is one
   */
  iterator begin () const
  {
    tl_assert (m_type == t_list);
    return m_var.m_list->begin ();
  }

  /**
   *  @brief Get the list iterators, if it is one
   */
  iterator end () const
  {
    tl_assert (m_type == t_list);
    return m_var.m_list->end ();
  }

  /**
   *  @brief Add a element to the list
   */
  void push (const tl::Variant &v)
  {
    tl_assert (m_type == t_list);
    m_var.m_list->push_back (v);
  }

  /**
   *  @brief Get the list, if it is one
   */
  std::vector<tl::Variant> &get_list ()
  {
    tl_assert (m_type == t_list);
    return *m_var.m_list;
  }

  /**
   *  @brief Test, if it is a double
   */
  bool is_double () const
  {
    return m_type == t_double;
  }

  /**
   *  @brief Test, if it is a long
   */
  bool is_long () const
  {
    return m_type == t_long;
  }

  /**
   *  @brief Test, if it is a id
   */
  bool is_id () const
  {
    return m_type == t_id;
  }

  /**
   *  @brief Test, if it is a string
   */
  bool is_string () const
  {
    return m_type == t_string;
  }

  /**
   *  @brief Test, if it is nil
   */
  bool is_nil () const
  {
    return m_type == t_nil;
  }

  /**
   *  @brief Test, if it is a list
   */
  bool is_list () const
  {
    return m_type == t_list;
  }

  /**
   *  @brief Test, if this is a user type
   */
  bool is_user () const
  {
    return m_type >= t_user;
  }

  /**
   *  @brief Get the user type if it is one
   */
  unsigned int user_type () const
  {
    tl_assert (is_user ());
    return (unsigned int) m_type - (unsigned int) t_user;
  }

  /**
   *  @brief A method to deliver an empty-list variant
   */
  static tl::Variant empty_list ();

  /**
   *  @brief Convert the Variant to a string that can be parsed with the Extractor
   *
   *  No conversion is provided for user types and "nil" currently.
   */
  std::string to_parsable_string () const;

private:
  type m_type;

  union {
    std::vector<tl::Variant> *m_list;
    double m_double;
    long m_long;
    VariantUserBase *mp_user;
  } m_var;

  //  this will hold the string if it is valid
  mutable char *m_string;

  void reset ();
};

/**
 *  @brief Special extractors for the variant
 */

class Extractor;
template <class T> void extractor_impl (tl::Extractor &ex, T &);
template <class T> bool test_extractor_impl (tl::Extractor &ex, T &);

template <> void extractor_impl (tl::Extractor &ex, tl::Variant &p);
template <> bool test_extractor_impl (tl::Extractor &ex, tl::Variant &p);

} // namespace tl

#endif
